#include <arpa/inet.h>

#include "reorderer.h"
#include "protocol.h"

#define PACKET_BUFFER_SIZE 100
#define TIMEOUT_SEC 0.100

using namespace std;

double tdiff(const timeval& a, const timeval& b)
{
	return b.tv_sec - a.tv_sec +
		1e-6 * (b.tv_usec - a.tv_usec);
}

Reorderer::Reorderer(Sender* sender)
	: sender(sender), last_seq(-1)
{
	gettimeofday(&last_sent_packet, NULL);
}

void Reorderer::send_packet(uint16_t proto, const string& data, int seq)
{
	timeval now;
	gettimeofday(&now, NULL);

	bool silence = false;
	if (packet_buffer.size() >= PACKET_BUFFER_SIZE) {
		printf("Gave up waiting for packets [%d,%d> (buffer full)\n",
			last_seq + 1, packet_buffer.top().seq);
		silence = true;
		last_seq = packet_buffer.top().seq - 1;
	} else if (!packet_buffer.empty() &&
	           tdiff(packet_buffer.top().ts, now) > TIMEOUT_SEC) {
		printf("Gave up waiting for packets [%d,%d> (timeout)\n",
			last_seq + 1, packet_buffer.top().seq);
		silence = true;
		last_seq = packet_buffer.top().seq - 1;
		// TODO: Rerun immediately after we've cleared out,
		// in case there are more timeouts.
	}

	// In case of restarts.
	if (packet_buffer.empty() &&
	    seq < last_seq &&
	    tdiff(last_sent_packet, now) > 5.0) {
		printf("No good data for five seconds, resetting sequence to %d\n", seq);
		last_seq = seq - 1;
	}

	GREPacket packet;
	packet.seq = seq;
	packet.proto = proto;
	packet.data = data;
	packet.ts = now;
	packet_buffer.push(packet);

	while (!packet_buffer.empty() &&
	       (last_seq == -1 || packet_buffer.top().seq <= last_seq + 1)) {
		int front_seq = packet_buffer.top().seq;
		if (front_seq < last_seq + 1) {
			printf("Duplicate packet or way out-of-order: seq=%d front_seq=%d\n",
				front_seq, last_seq + 1);
			packet_buffer.pop();
			continue;
		}
		//if (packet_buffer.size() > 1) {
		//	printf("seq=%d (REORDER %d)\n", front_seq, int(packet_buffer.size()));
		//} else {
		//	printf("seq=%d\n", front_seq);
		//}
		const string &data = packet_buffer.top().data;
		check_ts_discontinuity(packet_buffer.top().proto, data, silence);
		sender->send_packet(proto, data, seq);
		packet_buffer.pop();
		last_seq = front_seq;
		last_sent_packet = now;
		if (!silence && !packet_buffer.empty()) {
			printf("Reordering with packet buffer size %d: seq=%d new_front_seq=%d\n", int(packet_buffer.size()), front_seq, packet_buffer.top().seq);
			silence = true;
		}
	}
}

void Reorderer::check_ts_discontinuity(uint16_t proto, const string &data, bool silence)
{
	if (data.size() == 1344) {
		for (int i = 0; i < 7; ++i) {
			const char *pkt = &data[i * 188 + 28];
			int pid = (ntohl(*(uint32_t *)(pkt)) & 0x1fff00) >> 8;
			if (pid == 8191) {
				// stuffing, ignore
				continue;
			}
			int has_payload = pkt[3] & 0x10;
			int cc = pkt[3] & 0xf;
			if (has_payload) {
				int last_cc = ccs[pid];
				if (!silence && cc != ((last_cc + 1) & 0xf)) {
					printf("Pid %d discontinuity (expected %d, got %d)\n", pid, (last_cc + 1) & 0xf, cc);
				}
				ccs[pid] = cc;
			}
		}
	}
}

